default_config = {
    'char_match_threshold': 0.5,
    'ngram_threshold': 0.5,
    'skipgram_threshold': 0.3,
    'levenshtein_threshold': 0.5,
    'include_variants': False,
    'filter_distractors': False
}